// clang-format off
// Auto-generated by Build/gen_devirt.py. DO NOT EDIT -- regenerate instead.
//
// NEON-only RTCD devirtualization redirects (RTC guaranteed-Neon build).
// Each macro points a dispatch name at its _neon impl so callers emit a
// direct, inlinable call instead of an indirect function-pointer branch.
// The _neon symbols are already declared in the RTCD header (ARCH_AARCH64).
// Included under "#if CONFIG_ARM_NEON_IS_GUARANTEED && !defined(RTCD_C)".

#define svt_aom_blend_a64_mask svt_aom_blend_a64_mask_neon
#define svt_aom_blend_a64_hmask svt_aom_blend_a64_hmask_neon
#define svt_aom_blend_a64_vmask svt_aom_blend_a64_vmask_neon
#define svt_aom_lowbd_blend_a64_d16_mask svt_aom_lowbd_blend_a64_d16_mask_neon
#define svt_cfl_predict_lbd svt_aom_cfl_predict_lbd_neon
#define svt_cfl_predict_hbd svt_cfl_predict_hbd_neon
#define svt_av1_filter_intra_edge svt_av1_filter_intra_edge_neon
#define svt_av1_upsample_intra_edge svt_av1_upsample_intra_edge_neon
#define svt_av1_build_compound_diffwtd_mask_d16 svt_av1_build_compound_diffwtd_mask_d16_neon
#define svt_apply_selfguided_restoration svt_aom_apply_selfguided_restoration_neon
#define svt_av1_selfguided_restoration svt_av1_selfguided_restoration_neon
#define svt_av1_inv_txfm2d_add_4x4 svt_av1_inv_txfm2d_add_4x4_neon
#define svt_av1_inv_txfm2d_add_4x8 svt_av1_inv_txfm2d_add_4x8_neon
#define svt_av1_inv_txfm2d_add_4x16 svt_av1_inv_txfm2d_add_4x16_neon
#define svt_av1_inv_txfm2d_add_8x4 svt_av1_inv_txfm2d_add_8x4_neon
#define svt_av1_inv_txfm2d_add_8x8 svt_av1_inv_txfm2d_add_8x8_neon
#define svt_av1_inv_txfm2d_add_8x16 svt_av1_inv_txfm2d_add_8x16_neon
#define svt_av1_inv_txfm2d_add_8x32 svt_av1_inv_txfm2d_add_8x32_neon
#define svt_av1_inv_txfm2d_add_16x4 svt_av1_inv_txfm2d_add_16x4_neon
#define svt_av1_inv_txfm2d_add_16x8 svt_av1_inv_txfm2d_add_16x8_neon
#define svt_av1_inv_txfm2d_add_16x16 svt_av1_inv_txfm2d_add_16x16_neon
#define svt_av1_inv_txfm2d_add_16x32 svt_av1_inv_txfm2d_add_16x32_neon
#define svt_av1_inv_txfm2d_add_16x64 svt_av1_inv_txfm2d_add_16x64_neon
#define svt_av1_inv_txfm2d_add_32x8 svt_av1_inv_txfm2d_add_32x8_neon
#define svt_av1_inv_txfm2d_add_32x16 svt_av1_inv_txfm2d_add_32x16_neon
#define svt_av1_inv_txfm2d_add_32x32 svt_av1_inv_txfm2d_add_32x32_neon
#define svt_av1_inv_txfm2d_add_32x64 svt_av1_inv_txfm2d_add_32x64_neon
#define svt_av1_inv_txfm2d_add_64x16 svt_av1_inv_txfm2d_add_64x16_neon
#define svt_av1_inv_txfm2d_add_64x32 svt_av1_inv_txfm2d_add_64x32_neon
#define svt_av1_inv_txfm2d_add_64x64 svt_av1_inv_txfm2d_add_64x64_neon
#define svt_av1_inv_txfm_add svt_dav1d_inv_txfm_add_neon
#define svt_compressed_packmsb svt_compressed_packmsb_neon
#define svt_cfl_luma_subsampling_420_lbd svt_cfl_luma_subsampling_420_lbd_neon
#define svt_cfl_luma_subsampling_420_hbd svt_cfl_luma_subsampling_420_hbd_neon
#define svt_pack2d_16_bit_src_mul4 svt_enc_msb_pack2d_neon
#define svt_aom_un_pack2d_16_bit_src_mul4 svt_enc_msb_un_pack2d_neon
#define svt_full_distortion_kernel_cbf_zero32_bits svt_full_distortion_kernel_cbf_zero32_bits_neon
#define svt_full_distortion_kernel32_bits svt_full_distortion_kernel32_bits_neon
#define svt_residual_kernel8bit svt_residual_kernel8bit_neon
#define svt_residual_kernel16bit svt_residual_kernel16bit_neon
#define svt_av1_wiener_convolve_add_src svt_av1_wiener_convolve_add_src_neon
#define svt_av1_convolve_2d_copy_sr svt_av1_convolve_2d_copy_sr_neon
#define svt_av1_jnt_convolve_2d_copy svt_av1_jnt_convolve_2d_copy_neon
#define svt_av1_build_compound_diffwtd_mask svt_av1_build_compound_diffwtd_mask_neon
#define svt_aom_subtract_block svt_aom_subtract_block_neon
#define svt_av1_dr_prediction_z1 svt_av1_dr_prediction_z1_neon
#define svt_av1_dr_prediction_z2 svt_av1_dr_prediction_z2_neon
#define svt_av1_dr_prediction_z3 svt_av1_dr_prediction_z3_neon
#define svt_aom_paeth_predictor_4x4 svt_aom_paeth_predictor_4x4_neon
#define svt_aom_paeth_predictor_4x8 svt_aom_paeth_predictor_4x8_neon
#define svt_aom_paeth_predictor_4x16 svt_aom_paeth_predictor_4x16_neon
#define svt_aom_paeth_predictor_8x4 svt_aom_paeth_predictor_8x4_neon
#define svt_aom_paeth_predictor_8x8 svt_aom_paeth_predictor_8x8_neon
#define svt_aom_paeth_predictor_8x16 svt_aom_paeth_predictor_8x16_neon
#define svt_aom_paeth_predictor_8x32 svt_aom_paeth_predictor_8x32_neon
#define svt_aom_paeth_predictor_16x4 svt_aom_paeth_predictor_16x4_neon
#define svt_aom_paeth_predictor_16x8 svt_aom_paeth_predictor_16x8_neon
#define svt_aom_paeth_predictor_16x16 svt_aom_paeth_predictor_16x16_neon
#define svt_aom_paeth_predictor_16x32 svt_aom_paeth_predictor_16x32_neon
#define svt_aom_paeth_predictor_16x64 svt_aom_paeth_predictor_16x64_neon
#define svt_aom_paeth_predictor_32x8 svt_aom_paeth_predictor_32x8_neon
#define svt_aom_paeth_predictor_32x16 svt_aom_paeth_predictor_32x16_neon
#define svt_aom_paeth_predictor_32x32 svt_aom_paeth_predictor_32x32_neon
#define svt_aom_paeth_predictor_32x64 svt_aom_paeth_predictor_32x64_neon
#define svt_aom_paeth_predictor_64x16 svt_aom_paeth_predictor_64x16_neon
#define svt_aom_paeth_predictor_64x32 svt_aom_paeth_predictor_64x32_neon
#define svt_aom_paeth_predictor_64x64 svt_aom_paeth_predictor_64x64_neon
#define svt_aom_dc_predictor_4x4 svt_aom_dc_predictor_4x4_neon
#define svt_aom_dc_predictor_4x8 svt_aom_dc_predictor_4x8_neon
#define svt_aom_dc_predictor_4x16 svt_aom_dc_predictor_4x16_neon
#define svt_aom_dc_predictor_8x4 svt_aom_dc_predictor_8x4_neon
#define svt_aom_dc_predictor_8x8 svt_aom_dc_predictor_8x8_neon
#define svt_aom_dc_predictor_8x16 svt_aom_dc_predictor_8x16_neon
#define svt_aom_dc_predictor_8x32 svt_aom_dc_predictor_8x32_neon
#define svt_aom_dc_predictor_16x4 svt_aom_dc_predictor_16x4_neon
#define svt_aom_dc_predictor_16x8 svt_aom_dc_predictor_16x8_neon
#define svt_aom_dc_predictor_16x16 svt_aom_dc_predictor_16x16_neon
#define svt_aom_dc_predictor_16x32 svt_aom_dc_predictor_16x32_neon
#define svt_aom_dc_predictor_16x64 svt_aom_dc_predictor_16x64_neon
#define svt_aom_dc_predictor_32x8 svt_aom_dc_predictor_32x8_neon
#define svt_aom_dc_predictor_32x16 svt_aom_dc_predictor_32x16_neon
#define svt_aom_dc_predictor_32x32 svt_aom_dc_predictor_32x32_neon
#define svt_aom_dc_predictor_32x64 svt_aom_dc_predictor_32x64_neon
#define svt_aom_dc_predictor_64x16 svt_aom_dc_predictor_64x16_neon
#define svt_aom_dc_predictor_64x32 svt_aom_dc_predictor_64x32_neon
#define svt_aom_dc_predictor_64x64 svt_aom_dc_predictor_64x64_neon
#define svt_aom_dc_top_predictor_4x4 svt_aom_dc_top_predictor_4x4_neon
#define svt_aom_dc_top_predictor_4x8 svt_aom_dc_top_predictor_4x8_neon
#define svt_aom_dc_top_predictor_4x16 svt_aom_dc_top_predictor_4x16_neon
#define svt_aom_dc_top_predictor_8x4 svt_aom_dc_top_predictor_8x4_neon
#define svt_aom_dc_top_predictor_8x8 svt_aom_dc_top_predictor_8x8_neon
#define svt_aom_dc_top_predictor_8x16 svt_aom_dc_top_predictor_8x16_neon
#define svt_aom_dc_top_predictor_8x32 svt_aom_dc_top_predictor_8x32_neon
#define svt_aom_dc_top_predictor_16x4 svt_aom_dc_top_predictor_16x4_neon
#define svt_aom_dc_top_predictor_16x8 svt_aom_dc_top_predictor_16x8_neon
#define svt_aom_dc_top_predictor_16x16 svt_aom_dc_top_predictor_16x16_neon
#define svt_aom_dc_top_predictor_16x32 svt_aom_dc_top_predictor_16x32_neon
#define svt_aom_dc_top_predictor_16x64 svt_aom_dc_top_predictor_16x64_neon
#define svt_aom_dc_top_predictor_32x8 svt_aom_dc_top_predictor_32x8_neon
#define svt_aom_dc_top_predictor_32x16 svt_aom_dc_top_predictor_32x16_neon
#define svt_aom_dc_top_predictor_32x32 svt_aom_dc_top_predictor_32x32_neon
#define svt_aom_dc_top_predictor_32x64 svt_aom_dc_top_predictor_32x64_neon
#define svt_aom_dc_top_predictor_64x16 svt_aom_dc_top_predictor_64x16_neon
#define svt_aom_dc_top_predictor_64x32 svt_aom_dc_top_predictor_64x32_neon
#define svt_aom_dc_top_predictor_64x64 svt_aom_dc_top_predictor_64x64_neon
#define svt_aom_dc_left_predictor_4x4 svt_aom_dc_left_predictor_4x4_neon
#define svt_aom_dc_left_predictor_4x8 svt_aom_dc_left_predictor_4x8_neon
#define svt_aom_dc_left_predictor_4x16 svt_aom_dc_left_predictor_4x16_neon
#define svt_aom_dc_left_predictor_8x4 svt_aom_dc_left_predictor_8x4_neon
#define svt_aom_dc_left_predictor_8x8 svt_aom_dc_left_predictor_8x8_neon
#define svt_aom_dc_left_predictor_8x16 svt_aom_dc_left_predictor_8x16_neon
#define svt_aom_dc_left_predictor_8x32 svt_aom_dc_left_predictor_8x32_neon
#define svt_aom_dc_left_predictor_16x4 svt_aom_dc_left_predictor_16x4_neon
#define svt_aom_dc_left_predictor_16x8 svt_aom_dc_left_predictor_16x8_neon
#define svt_aom_dc_left_predictor_16x16 svt_aom_dc_left_predictor_16x16_neon
#define svt_aom_dc_left_predictor_16x32 svt_aom_dc_left_predictor_16x32_neon
#define svt_aom_dc_left_predictor_16x64 svt_aom_dc_left_predictor_16x64_neon
#define svt_aom_dc_left_predictor_32x8 svt_aom_dc_left_predictor_32x8_neon
#define svt_aom_dc_left_predictor_32x16 svt_aom_dc_left_predictor_32x16_neon
#define svt_aom_dc_left_predictor_32x32 svt_aom_dc_left_predictor_32x32_neon
#define svt_aom_dc_left_predictor_32x64 svt_aom_dc_left_predictor_32x64_neon
#define svt_aom_dc_left_predictor_64x16 svt_aom_dc_left_predictor_64x16_neon
#define svt_aom_dc_left_predictor_64x32 svt_aom_dc_left_predictor_64x32_neon
#define svt_aom_dc_left_predictor_64x64 svt_aom_dc_left_predictor_64x64_neon
#define svt_aom_dc_128_predictor_4x4 svt_aom_dc_128_predictor_4x4_neon
#define svt_aom_dc_128_predictor_4x8 svt_aom_dc_128_predictor_4x8_neon
#define svt_aom_dc_128_predictor_4x16 svt_aom_dc_128_predictor_4x16_neon
#define svt_aom_dc_128_predictor_8x4 svt_aom_dc_128_predictor_8x4_neon
#define svt_aom_dc_128_predictor_8x8 svt_aom_dc_128_predictor_8x8_neon
#define svt_aom_dc_128_predictor_8x16 svt_aom_dc_128_predictor_8x16_neon
#define svt_aom_dc_128_predictor_8x32 svt_aom_dc_128_predictor_8x32_neon
#define svt_aom_dc_128_predictor_16x4 svt_aom_dc_128_predictor_16x4_neon
#define svt_aom_dc_128_predictor_16x8 svt_aom_dc_128_predictor_16x8_neon
#define svt_aom_dc_128_predictor_16x16 svt_aom_dc_128_predictor_16x16_neon
#define svt_aom_dc_128_predictor_16x32 svt_aom_dc_128_predictor_16x32_neon
#define svt_aom_dc_128_predictor_16x64 svt_aom_dc_128_predictor_16x64_neon
#define svt_aom_dc_128_predictor_32x8 svt_aom_dc_128_predictor_32x8_neon
#define svt_aom_dc_128_predictor_32x16 svt_aom_dc_128_predictor_32x16_neon
#define svt_aom_dc_128_predictor_32x32 svt_aom_dc_128_predictor_32x32_neon
#define svt_aom_dc_128_predictor_32x64 svt_aom_dc_128_predictor_32x64_neon
#define svt_aom_dc_128_predictor_64x16 svt_aom_dc_128_predictor_64x16_neon
#define svt_aom_dc_128_predictor_64x32 svt_aom_dc_128_predictor_64x32_neon
#define svt_aom_dc_128_predictor_64x64 svt_aom_dc_128_predictor_64x64_neon
#define svt_aom_smooth_h_predictor_4x4 svt_aom_smooth_h_predictor_4x4_neon
#define svt_aom_smooth_h_predictor_4x8 svt_aom_smooth_h_predictor_4x8_neon
#define svt_aom_smooth_h_predictor_4x16 svt_aom_smooth_h_predictor_4x16_neon
#define svt_aom_smooth_h_predictor_8x4 svt_aom_smooth_h_predictor_8x4_neon
#define svt_aom_smooth_h_predictor_8x8 svt_aom_smooth_h_predictor_8x8_neon
#define svt_aom_smooth_h_predictor_8x16 svt_aom_smooth_h_predictor_8x16_neon
#define svt_aom_smooth_h_predictor_8x32 svt_aom_smooth_h_predictor_8x32_neon
#define svt_aom_smooth_h_predictor_16x4 svt_aom_smooth_h_predictor_16x4_neon
#define svt_aom_smooth_h_predictor_16x8 svt_aom_smooth_h_predictor_16x8_neon
#define svt_aom_smooth_h_predictor_16x16 svt_aom_smooth_h_predictor_16x16_neon
#define svt_aom_smooth_h_predictor_16x32 svt_aom_smooth_h_predictor_16x32_neon
#define svt_aom_smooth_h_predictor_16x64 svt_aom_smooth_h_predictor_16x64_neon
#define svt_aom_smooth_h_predictor_32x8 svt_aom_smooth_h_predictor_32x8_neon
#define svt_aom_smooth_h_predictor_32x16 svt_aom_smooth_h_predictor_32x16_neon
#define svt_aom_smooth_h_predictor_32x32 svt_aom_smooth_h_predictor_32x32_neon
#define svt_aom_smooth_h_predictor_32x64 svt_aom_smooth_h_predictor_32x64_neon
#define svt_aom_smooth_h_predictor_64x16 svt_aom_smooth_h_predictor_64x16_neon
#define svt_aom_smooth_h_predictor_64x32 svt_aom_smooth_h_predictor_64x32_neon
#define svt_aom_smooth_h_predictor_64x64 svt_aom_smooth_h_predictor_64x64_neon
#define svt_aom_smooth_v_predictor_4x4 svt_aom_smooth_v_predictor_4x4_neon
#define svt_aom_smooth_v_predictor_4x8 svt_aom_smooth_v_predictor_4x8_neon
#define svt_aom_smooth_v_predictor_4x16 svt_aom_smooth_v_predictor_4x16_neon
#define svt_aom_smooth_v_predictor_8x4 svt_aom_smooth_v_predictor_8x4_neon
#define svt_aom_smooth_v_predictor_8x8 svt_aom_smooth_v_predictor_8x8_neon
#define svt_aom_smooth_v_predictor_8x16 svt_aom_smooth_v_predictor_8x16_neon
#define svt_aom_smooth_v_predictor_8x32 svt_aom_smooth_v_predictor_8x32_neon
#define svt_aom_smooth_v_predictor_16x4 svt_aom_smooth_v_predictor_16x4_neon
#define svt_aom_smooth_v_predictor_16x8 svt_aom_smooth_v_predictor_16x8_neon
#define svt_aom_smooth_v_predictor_16x16 svt_aom_smooth_v_predictor_16x16_neon
#define svt_aom_smooth_v_predictor_16x32 svt_aom_smooth_v_predictor_16x32_neon
#define svt_aom_smooth_v_predictor_16x64 svt_aom_smooth_v_predictor_16x64_neon
#define svt_aom_smooth_v_predictor_32x8 svt_aom_smooth_v_predictor_32x8_neon
#define svt_aom_smooth_v_predictor_32x16 svt_aom_smooth_v_predictor_32x16_neon
#define svt_aom_smooth_v_predictor_32x32 svt_aom_smooth_v_predictor_32x32_neon
#define svt_aom_smooth_v_predictor_32x64 svt_aom_smooth_v_predictor_32x64_neon
#define svt_aom_smooth_v_predictor_64x16 svt_aom_smooth_v_predictor_64x16_neon
#define svt_aom_smooth_v_predictor_64x32 svt_aom_smooth_v_predictor_64x32_neon
#define svt_aom_smooth_v_predictor_64x64 svt_aom_smooth_v_predictor_64x64_neon
#define svt_aom_smooth_predictor_4x4 svt_aom_smooth_predictor_4x4_neon
#define svt_aom_smooth_predictor_4x8 svt_aom_smooth_predictor_4x8_neon
#define svt_aom_smooth_predictor_4x16 svt_aom_smooth_predictor_4x16_neon
#define svt_aom_smooth_predictor_8x4 svt_aom_smooth_predictor_8x4_neon
#define svt_aom_smooth_predictor_8x8 svt_aom_smooth_predictor_8x8_neon
#define svt_aom_smooth_predictor_8x16 svt_aom_smooth_predictor_8x16_neon
#define svt_aom_smooth_predictor_8x32 svt_aom_smooth_predictor_8x32_neon
#define svt_aom_smooth_predictor_16x4 svt_aom_smooth_predictor_16x4_neon
#define svt_aom_smooth_predictor_16x8 svt_aom_smooth_predictor_16x8_neon
#define svt_aom_smooth_predictor_16x16 svt_aom_smooth_predictor_16x16_neon
#define svt_aom_smooth_predictor_16x32 svt_aom_smooth_predictor_16x32_neon
#define svt_aom_smooth_predictor_16x64 svt_aom_smooth_predictor_16x64_neon
#define svt_aom_smooth_predictor_32x8 svt_aom_smooth_predictor_32x8_neon
#define svt_aom_smooth_predictor_32x16 svt_aom_smooth_predictor_32x16_neon
#define svt_aom_smooth_predictor_32x32 svt_aom_smooth_predictor_32x32_neon
#define svt_aom_smooth_predictor_32x64 svt_aom_smooth_predictor_32x64_neon
#define svt_aom_smooth_predictor_64x16 svt_aom_smooth_predictor_64x16_neon
#define svt_aom_smooth_predictor_64x32 svt_aom_smooth_predictor_64x32_neon
#define svt_aom_smooth_predictor_64x64 svt_aom_smooth_predictor_64x64_neon
#define svt_aom_v_predictor_4x4 svt_aom_v_predictor_4x4_neon
#define svt_aom_v_predictor_4x8 svt_aom_v_predictor_4x8_neon
#define svt_aom_v_predictor_4x16 svt_aom_v_predictor_4x16_neon
#define svt_aom_v_predictor_8x4 svt_aom_v_predictor_8x4_neon
#define svt_aom_v_predictor_8x8 svt_aom_v_predictor_8x8_neon
#define svt_aom_v_predictor_8x16 svt_aom_v_predictor_8x16_neon
#define svt_aom_v_predictor_8x32 svt_aom_v_predictor_8x32_neon
#define svt_aom_v_predictor_16x4 svt_aom_v_predictor_16x4_neon
#define svt_aom_v_predictor_16x8 svt_aom_v_predictor_16x8_neon
#define svt_aom_v_predictor_16x16 svt_aom_v_predictor_16x16_neon
#define svt_aom_v_predictor_16x32 svt_aom_v_predictor_16x32_neon
#define svt_aom_v_predictor_16x64 svt_aom_v_predictor_16x64_neon
#define svt_aom_v_predictor_32x8 svt_aom_v_predictor_32x8_neon
#define svt_aom_v_predictor_32x16 svt_aom_v_predictor_32x16_neon
#define svt_aom_v_predictor_32x32 svt_aom_v_predictor_32x32_neon
#define svt_aom_v_predictor_32x64 svt_aom_v_predictor_32x64_neon
#define svt_aom_v_predictor_64x16 svt_aom_v_predictor_64x16_neon
#define svt_aom_v_predictor_64x32 svt_aom_v_predictor_64x32_neon
#define svt_aom_v_predictor_64x64 svt_aom_v_predictor_64x64_neon
#define svt_aom_h_predictor_4x4 svt_aom_h_predictor_4x4_neon
#define svt_aom_h_predictor_4x8 svt_aom_h_predictor_4x8_neon
#define svt_aom_h_predictor_4x16 svt_aom_h_predictor_4x16_neon
#define svt_aom_h_predictor_8x4 svt_aom_h_predictor_8x4_neon
#define svt_aom_h_predictor_8x8 svt_aom_h_predictor_8x8_neon
#define svt_aom_h_predictor_8x16 svt_aom_h_predictor_8x16_neon
#define svt_aom_h_predictor_8x32 svt_aom_h_predictor_8x32_neon
#define svt_aom_h_predictor_16x4 svt_aom_h_predictor_16x4_neon
#define svt_aom_h_predictor_16x8 svt_aom_h_predictor_16x8_neon
#define svt_aom_h_predictor_16x16 svt_aom_h_predictor_16x16_neon
#define svt_aom_h_predictor_16x32 svt_aom_h_predictor_16x32_neon
#define svt_aom_h_predictor_16x64 svt_aom_h_predictor_16x64_neon
#define svt_aom_h_predictor_32x8 svt_aom_h_predictor_32x8_neon
#define svt_aom_h_predictor_32x16 svt_aom_h_predictor_32x16_neon
#define svt_aom_h_predictor_32x32 svt_aom_h_predictor_32x32_neon
#define svt_aom_h_predictor_32x64 svt_aom_h_predictor_32x64_neon
#define svt_aom_h_predictor_64x16 svt_aom_h_predictor_64x16_neon
#define svt_aom_h_predictor_64x32 svt_aom_h_predictor_64x32_neon
#define svt_aom_h_predictor_64x64 svt_aom_h_predictor_64x64_neon
#define svt_aom_cdef_find_dir svt_aom_cdef_find_dir_neon
#define svt_aom_cdef_find_dir_dual svt_aom_cdef_find_dir_dual_neon
#define svt_cdef_filter_block svt_cdef_filter_block_neon
#define svt_cdef_filter_block_8bit svt_cdef_filter_block_8bit_neon
#define svt_cdef_filter_block_8bit_bounded svt_cdef_filter_block_8bit_bounded_neon
#define svt_aom_cdef_find_dir_8bit svt_aom_cdef_find_dir_8bit_neon
#define svt_aom_cdef_find_dir_dual_8bit svt_aom_cdef_find_dir_dual_8bit_neon
#define svt_aom_copy_rect8_8bit_to_16bit svt_aom_copy_rect8_8bit_to_16bit_neon
#define svt_aom_lpf_horizontal_4 svt_aom_lpf_horizontal_4_neon
#define svt_aom_lpf_horizontal_6 svt_aom_lpf_horizontal_6_neon
#define svt_aom_lpf_horizontal_8 svt_aom_lpf_horizontal_8_neon
#define svt_aom_lpf_horizontal_14 svt_aom_lpf_horizontal_14_neon
#define svt_aom_lpf_vertical_4 svt_aom_lpf_vertical_4_neon
#define svt_aom_lpf_vertical_6 svt_aom_lpf_vertical_6_neon
#define svt_aom_lpf_vertical_8 svt_aom_lpf_vertical_8_neon
#define svt_aom_lpf_vertical_14 svt_aom_lpf_vertical_14_neon
#define svt_av1_copy_wxh_8bit svt_av1_copy_wxh_8bit_neon
#define svt_av1_copy_wxh_16bit svt_av1_copy_wxh_16bit_neon
#define svt_memcpy svt_memcpy_neon
#define svt_memset svt_memset_neon
#define svt_aom_hadamard_32x32 svt_aom_hadamard_32x32_neon
#define svt_aom_hadamard_16x16 svt_aom_hadamard_16x16_neon
#define svt_aom_hadamard_8x8 svt_aom_hadamard_8x8_neon
#define svt_aom_hadamard_4x4 svt_aom_hadamard_4x4_neon
// clang-format on
