// clang-format off
// Auto-generated by Build/gen_devirt.py. DO NOT EDIT -- regenerate instead.
//
// NEON-only RTCD devirtualization redirects (RTC guaranteed-Neon build).
// Each macro points a dispatch name at its _neon impl so callers emit a
// direct, inlinable call instead of an indirect function-pointer branch.
// The _neon symbols are already declared in the RTCD header (ARCH_AARCH64).
// Included under "#if CONFIG_ARM_NEON_IS_GUARANTEED && !defined(AOM_RTCD_C)".

#define svt_av1_wedge_compute_delta_squares svt_av1_wedge_compute_delta_squares_neon
#define svt_subtract_average svt_subtract_average_neon
#define svt_get_proj_subspace svt_get_proj_subspace_neon
#define svt_aom_quantize_b svt_aom_quantize_b_neon
#define svt_av1_quantize_b_qm svt_av1_quantize_b_qm_neon
#define svt_av1_quantize_fp svt_av1_quantize_fp_neon
#define svt_av1_quantize_fp_32x32 svt_av1_quantize_fp_32x32_neon
#define svt_av1_quantize_fp_64x64 svt_av1_quantize_fp_64x64_neon
#define svt_av1_quantize_fp_qm svt_av1_quantize_fp_qm_neon
#define svt_aom_sad4x4 svt_aom_sad4x4_neon
#define svt_aom_sad4x4x4d svt_aom_sad4x4x4d_neon
#define svt_aom_sad4x16 svt_aom_sad4x16_neon
#define svt_aom_sad4x16x4d svt_aom_sad4x16x4d_neon
#define svt_aom_sad4x8 svt_aom_sad4x8_neon
#define svt_aom_sad4x8x4d svt_aom_sad4x8x4d_neon
#define svt_aom_sad64x128x4d svt_aom_sad64x128x4d_neon
#define svt_aom_sad64x16x4d svt_aom_sad64x16x4d_neon
#define svt_aom_sad64x32x4d svt_aom_sad64x32x4d_neon
#define svt_aom_sad64x64x4d svt_aom_sad64x64x4d_neon
#define svt_aom_sad8x16 svt_aom_sad8x16_neon
#define svt_aom_sad8x16x4d svt_aom_sad8x16x4d_neon
#define svt_aom_sad8x32 svt_aom_sad8x32_neon
#define svt_aom_sad8x32x4d svt_aom_sad8x32x4d_neon
#define svt_aom_sad8x8 svt_aom_sad8x8_neon
#define svt_aom_sad8x8x4d svt_aom_sad8x8x4d_neon
#define svt_aom_sad16x4 svt_aom_sad16x4_neon
#define svt_aom_sad16x4x4d svt_aom_sad16x4x4d_neon
#define svt_aom_sad32x8 svt_aom_sad32x8_neon
#define svt_aom_sad32x8x4d svt_aom_sad32x8x4d_neon
#define svt_aom_sad16x64 svt_aom_sad16x64_neon
#define svt_aom_sad16x64x4d svt_aom_sad16x64x4d_neon
#define svt_aom_sad32x16 svt_aom_sad32x16_neon
#define svt_aom_sad32x16x4d svt_aom_sad32x16x4d_neon
#define svt_aom_sad16x32 svt_aom_sad16x32_neon
#define svt_aom_sad16x32x4d svt_aom_sad16x32x4d_neon
#define svt_aom_sad32x64 svt_aom_sad32x64_neon
#define svt_aom_sad32x64x4d svt_aom_sad32x64x4d_neon
#define svt_aom_sad32x32 svt_aom_sad32x32_neon
#define svt_aom_sad32x32x4d svt_aom_sad32x32x4d_neon
#define svt_aom_sad16x16 svt_aom_sad16x16_neon
#define svt_aom_sad16x16x4d svt_aom_sad16x16x4d_neon
#define svt_aom_sad16x8 svt_aom_sad16x8_neon
#define svt_aom_sad16x8x4d svt_aom_sad16x8x4d_neon
#define svt_aom_sad8x4 svt_aom_sad8x4_neon
#define svt_aom_sad8x4x4d svt_aom_sad8x4x4d_neon
#define svt_aom_sad64x16 svt_aom_sad64x16_neon
#define svt_aom_sad64x32 svt_aom_sad64x32_neon
#define svt_aom_sad64x64 svt_aom_sad64x64_neon
#define svt_aom_sad64x128 svt_aom_sad64x128_neon
#define svt_aom_sad128x128 svt_aom_sad128x128_neon
#define svt_aom_sad128x128x4d svt_aom_sad128x128x4d_neon
#define svt_aom_sad128x64 svt_aom_sad128x64_neon
#define svt_aom_sad128x64x4d svt_aom_sad128x64x4d_neon
#define svt_av1_txb_init_levels svt_av1_txb_init_levels_neon
#define svt_aom_satd svt_aom_satd_neon
#define svt_aom_upsampled_pred svt_aom_upsampled_pred_neon
#define svt_aom_variance4x4 svt_aom_variance4x4_neon
#define svt_aom_sub_pixel_variance4x4 svt_aom_sub_pixel_variance4x4_neon
#define svt_av1_fwd_txfm2d_4x4 svt_av1_fwd_txfm2d_4x4_neon
#define svt_av1_fwd_txfm2d_4x8 svt_av1_fwd_txfm2d_4x8_neon
#define svt_av1_fwd_txfm2d_4x16 svt_av1_fwd_txfm2d_4x16_neon
#define svt_av1_fwd_txfm2d_8x4 svt_av1_fwd_txfm2d_8x4_neon
#define svt_av1_fwd_txfm2d_8x8 svt_av1_fwd_txfm2d_8x8_neon
#define svt_av1_fwd_txfm2d_8x16 svt_av1_fwd_txfm2d_8x16_neon
#define svt_av1_fwd_txfm2d_8x32 svt_av1_fwd_txfm2d_8x32_neon
#define svt_av1_fwd_txfm2d_16x4 svt_av1_fwd_txfm2d_16x4_neon
#define svt_av1_fwd_txfm2d_16x8 svt_av1_fwd_txfm2d_16x8_neon
#define svt_av1_fwd_txfm2d_16x16 svt_av1_fwd_txfm2d_16x16_neon
#define svt_av1_fwd_txfm2d_16x32 svt_av1_fwd_txfm2d_16x32_neon
#define svt_av1_fwd_txfm2d_16x64 svt_av1_fwd_txfm2d_16x64_neon
#define svt_av1_fwd_txfm2d_32x8 svt_av1_fwd_txfm2d_32x8_neon
#define svt_av1_fwd_txfm2d_32x16 svt_av1_fwd_txfm2d_32x16_neon
#define svt_av1_fwd_txfm2d_32x32 svt_av1_fwd_txfm2d_32x32_neon
#define svt_av1_fwd_txfm2d_32x64 svt_av1_fwd_txfm2d_32x64_neon
#define svt_av1_fwd_txfm2d_64x16 svt_av1_fwd_txfm2d_64x16_neon
#define svt_av1_fwd_txfm2d_64x32 svt_av1_fwd_txfm2d_64x32_neon
#define svt_av1_fwd_txfm2d_64x64 svt_av1_fwd_txfm2d_64x64_neon
#define svt_handle_transform16x64 svt_handle_transform16x64_neon
#define svt_handle_transform32x64 svt_handle_transform32x64_neon
#define svt_handle_transform64x16 svt_handle_transform64x16_neon
#define svt_handle_transform64x32 svt_handle_transform64x32_neon
#define svt_handle_transform64x64 svt_handle_transform64x64_neon
#define svt_handle_transform16x64_N2_N4 svt_handle_transform16x64_N2_N4_neon
#define svt_handle_transform32x64_N2_N4 svt_handle_transform32x64_N2_N4_neon
#define svt_handle_transform64x16_N2_N4 svt_handle_transform64x16_N2_N4_neon
#define svt_handle_transform64x32_N2_N4 svt_handle_transform64x32_N2_N4_neon
#define svt_handle_transform64x64_N2_N4 svt_handle_transform64x64_N2_N4_neon
#define svt_av1_fwd_txfm2d_4x4_N2 svt_av1_fwd_txfm2d_4x4_N2_neon
#define svt_av1_fwd_txfm2d_4x8_N2 svt_av1_fwd_txfm2d_4x8_N2_neon
#define svt_av1_fwd_txfm2d_4x16_N2 svt_av1_fwd_txfm2d_4x16_N2_neon
#define svt_av1_fwd_txfm2d_8x4_N2 svt_av1_fwd_txfm2d_8x4_N2_neon
#define svt_av1_fwd_txfm2d_8x8_N2 svt_av1_fwd_txfm2d_8x8_N2_neon
#define svt_av1_fwd_txfm2d_8x16_N2 svt_av1_fwd_txfm2d_8x16_N2_neon
#define svt_av1_fwd_txfm2d_8x32_N2 svt_av1_fwd_txfm2d_8x32_N2_neon
#define svt_av1_fwd_txfm2d_16x4_N2 svt_av1_fwd_txfm2d_16x4_N2_neon
#define svt_av1_fwd_txfm2d_16x8_N2 svt_av1_fwd_txfm2d_16x8_N2_neon
#define svt_av1_fwd_txfm2d_16x16_N2 svt_av1_fwd_txfm2d_16x16_N2_neon
#define svt_av1_fwd_txfm2d_16x32_N2 svt_av1_fwd_txfm2d_16x32_N2_neon
#define svt_av1_fwd_txfm2d_16x64_N2 svt_av1_fwd_txfm2d_16x64_N2_neon
#define svt_av1_fwd_txfm2d_32x8_N2 svt_av1_fwd_txfm2d_32x8_N2_neon
#define svt_av1_fwd_txfm2d_32x16_N2 svt_av1_fwd_txfm2d_32x16_N2_neon
#define svt_av1_fwd_txfm2d_32x32_N2 svt_av1_fwd_txfm2d_32x32_N2_neon
#define svt_av1_fwd_txfm2d_32x64_N2 svt_av1_fwd_txfm2d_32x64_N2_neon
#define svt_av1_fwd_txfm2d_64x16_N2 svt_av1_fwd_txfm2d_64x16_N2_neon
#define svt_av1_fwd_txfm2d_64x32_N2 svt_av1_fwd_txfm2d_64x32_N2_neon
#define svt_av1_fwd_txfm2d_64x64_N2 svt_av1_fwd_txfm2d_64x64_N2_neon
#define svt_av1_fwd_txfm2d_4x4_N4 svt_av1_fwd_txfm2d_4x4_N4_neon
#define svt_av1_fwd_txfm2d_4x8_N4 svt_av1_fwd_txfm2d_4x8_N4_neon
#define svt_av1_fwd_txfm2d_4x16_N4 svt_av1_fwd_txfm2d_4x16_N4_neon
#define svt_av1_fwd_txfm2d_8x4_N4 svt_av1_fwd_txfm2d_8x4_N4_neon
#define svt_av1_fwd_txfm2d_8x8_N4 svt_av1_fwd_txfm2d_8x8_N4_neon
#define svt_av1_fwd_txfm2d_8x16_N4 svt_av1_fwd_txfm2d_8x16_N4_neon
#define svt_av1_fwd_txfm2d_8x32_N4 svt_av1_fwd_txfm2d_8x32_N4_neon
#define svt_av1_fwd_txfm2d_16x4_N4 svt_av1_fwd_txfm2d_16x4_N4_neon
#define svt_av1_fwd_txfm2d_16x8_N4 svt_av1_fwd_txfm2d_16x8_N4_neon
#define svt_av1_fwd_txfm2d_16x16_N4 svt_av1_fwd_txfm2d_16x16_N4_neon
#define svt_av1_fwd_txfm2d_16x32_N4 svt_av1_fwd_txfm2d_16x32_N4_neon
#define svt_av1_fwd_txfm2d_16x64_N4 svt_av1_fwd_txfm2d_16x64_N4_neon
#define svt_av1_fwd_txfm2d_32x8_N4 svt_av1_fwd_txfm2d_32x8_N4_neon
#define svt_av1_fwd_txfm2d_32x16_N4 svt_av1_fwd_txfm2d_32x16_N4_neon
#define svt_av1_fwd_txfm2d_32x32_N4 svt_av1_fwd_txfm2d_32x32_N4_neon
#define svt_av1_fwd_txfm2d_32x64_N4 svt_av1_fwd_txfm2d_32x64_N4_neon
#define svt_av1_fwd_txfm2d_64x16_N4 svt_av1_fwd_txfm2d_64x16_N4_neon
#define svt_av1_fwd_txfm2d_64x32_N4 svt_av1_fwd_txfm2d_64x32_N4_neon
#define svt_av1_fwd_txfm2d_64x64_N4 svt_av1_fwd_txfm2d_64x64_N4_neon
#define svt_av1_get_nz_map_contexts svt_av1_get_nz_map_contexts_neon
#define svt_search_one_dual svt_search_one_dual_neon
#define svt_pme_sad_loop_kernel svt_pme_sad_loop_kernel_neon
#define svt_av1_apply_zz_based_temporal_filter_planewise_medium svt_av1_apply_zz_based_temporal_filter_planewise_medium_neon
#define svt_av1_apply_zz_based_temporal_filter_planewise_medium_hbd svt_av1_apply_zz_based_temporal_filter_planewise_medium_hbd_neon
#define svt_av1_apply_temporal_filter_planewise_medium svt_av1_apply_temporal_filter_planewise_medium_neon
#define svt_av1_apply_temporal_filter_planewise_medium_hbd svt_av1_apply_temporal_filter_planewise_medium_hbd_neon
#define apply_filtering_central svt_aom_apply_filtering_central_neon
#define downsample_2d svt_aom_downsample_2d_neon
#define svt_ext_sad_calculation_32x32_64x64 svt_ext_sad_calculation_32x32_64x64_neon
#define svt_ext_all_sad_calculation_8x8_16x16 svt_ext_all_sad_calculation_8x8_16x16_neon
#define svt_ext_eight_sad_calculation_32x32_64x64 svt_ext_eight_sad_calculation_32x32_64x64_neon
#define svt_nxm_sad_kernel svt_nxm_sad_kernel_helper_neon
#define sad_16b_kernel svt_aom_sad_16b_kernel_neon
#define svt_av1_calc_indices_dim1 svt_av1_calc_indices_dim1_neon
#define svt_av1_calc_indices_dim2 svt_av1_calc_indices_dim2_neon
#define svt_unpack_and_2bcompress svt_unpack_and_2bcompress_neon
#define svt_estimate_noise_fp16 svt_estimate_noise_fp16_neon
#define svt_copy_mi_map_grid svt_copy_mi_map_grid_neon
#define svt_ssim_8x8_hbd svt_ssim_8x8_hbd_neon
#define svt_ssim_4x4_hbd svt_ssim_4x4_hbd_neon
// clang-format on
